import 'source-map-support/register';
import { relative, resolve } from 'path';
import { createDependencies, createDiagnosticMissingImport, createProgramPlugin, getDebug, getImportedName, isImportExport, isImportFromNodeModules, isImportNative, replaceImportExportSpecifier, resolveModule, resolveTypescriptModule, splitPackageName, testProjectFile, } from "@build-script/typescript-transformer-common/lib/_export_all_in_one_index.js";
import { error } from "fancy-log/index.js";
import { SyntaxKind, visitEachChild } from 'typescript';
export default createProgramPlugin(function plugin(program, { verbose } = {}, { packageJsonPath }) {
    const debug = getDebug(verbose || false);
    debug('load package.json: %s', packageJsonPath);
    const dependencies = createDependencies(packageJsonPath);
    debug('    dependencies count: %s', Object.keys(dependencies).length);
    return function transformer(transformationContext) {
        transformationContext.enableEmitNotification(SyntaxKind.SourceFile);
        function visitNode(node) {
            const fileName = node.getSourceFile().fileName;
            if (isImportExport(node)) {
                const importId = getImportedName(node);
                if (isImportNative(importId)) {
                    return node;
                }
                debug(' * %s', node.getText(node.getSourceFile()).split('\n').join(' '));
                if (isImportFromNodeModules(importId)) {
                    const { packageName } = splitPackageName(importId);
                    if (!dependencies[packageName]) {
                        error(createDiagnosticMissingImport(node));
                    }
                    else {
                        const info = resolveTypescriptModule(node, packageJsonPath);
                        if (info.type === 'missing') {
                            error(createDiagnosticMissingImport(node));
                        }
                        else if (info.identifiers.length > 0) {
                            debug('   -> replace to %s', info.nodeResolve);
                            return replaceImportExportSpecifier(node, info.nodeResolve);
                        }
                    }
                }
                else {
                    debug('   -> not imported');
                    // not node_module
                }
                if (testProjectFile(debug, fileName, node, program)) {
                    const importPath = node.moduleSpecifier.text;
                    return replaceImportExportSpecifier(node, importPath + '.js');
                }
            }
            return node;
        }
        return (sourceFile) => {
            const { fileName } = sourceFile;
            debug('[trans] visit \x1B[2m%s\x1B[0m', fileName);
            const result = visitEachChild(sourceFile, visitNode, transformationContext);
            debug('[trans] visit complete');
            return result;
        };
    };
});
export function transformExternalModule(debug, node, packageJsonPath) {
    const importPath = node.moduleSpecifier.text;
    const { packageName, filePath } = splitPackageName(importPath);
    const result = resolveModule('module', resolve(packageJsonPath, '../node_modules', packageName, 'package.json'), filePath, true);
    if (result) {
        debug('resolve module result:', result);
        const id = relative(resolve(packageJsonPath, '../node_modules'), result).replace(/^\.\//, '');
        return replaceImportExportSpecifier(node, id);
    }
    else {
        error(createDiagnosticMissingImport(node));
    }
    return node;
}
//# sourceMappingURL=transform.js.map